/*
 * bsc1195950_i915_gem
 *
 * Fix for CVE-2022-0330, bsc#1195950 (drivers/gpu/drm/i915/i915_gem.c part)
 *
 *  Copyright (c) 2022 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_DRM_I915)

#if !IS_MODULE(CONFIG_DRM_I915)
#error "Live patch supports only CONFIG_DRM_I915=m"
#endif

#if !IS_MODULE(CONFIG_DRM)
#error "Live patch supports only CONFIG_DRM=m"
#endif

/* klp-ccp: from drivers/gpu/drm/i915/i915_gem.c */
#include <drm/drmP.h>
#include <drm/drm_vma_manager.h>
#include <drm/i915_drm.h>

#include "bsc1195950_common.h"

/* klp-ccp: from drivers/gpu/drm/i915/i915_drv.h */
static void (*kgre_intel_uncore_forcewake_get)(struct drm_i915_private *dev_priv,
				enum forcewake_domains domains);
static void (*kgre_intel_uncore_forcewake_put)(struct drm_i915_private *dev_priv,
				enum forcewake_domains domains);

int kgrp_i915_gem_object_put_pages(struct drm_i915_gem_object *obj);

static bool (*kgre_i915_gem_obj_bound_any)(struct drm_i915_gem_object *o);

static void (*kgre_intel_runtime_pm_get)(struct drm_i915_private *dev_priv);
static void (*kgre_intel_runtime_pm_put)(struct drm_i915_private *dev_priv);

/* klp-ccp: from drivers/gpu/drm/i915/intel_ringbuffer.h */
static bool (*kgre_intel_ring_initialized)(struct intel_engine_cs *ring);

/* klp-ccp: from include/drm/drmP.h */
static __printf(1, 2)
void (*kgre_drm_err)(const char *format, ...);

#define KGRR_DRM_ERROR_RATELIMITED(fmt, ...)				\
({									\
	static DEFINE_RATELIMIT_STATE(_rs,				\
				      DEFAULT_RATELIMIT_INTERVAL,	\
				      DEFAULT_RATELIMIT_BURST);		\
									\
	if (__ratelimit(&_rs))						\
		(*kgre_drm_err)(fmt, ##__VA_ARGS__);			\
})

/* klp-ccp: from drivers/gpu/drm/i915/i915_gem.c */
#include <linux/slab.h>
#include <linux/swap.h>
#include <linux/pci.h>

static void
(*kgre_i915_gem_object_truncate)(struct drm_i915_gem_object *obj);

static void
kgrr_i915_gem_object_invalidate(struct drm_i915_gem_object *obj)
{
	struct address_space *mapping;

	switch (obj->madv) {
	case I915_MADV_DONTNEED:
		(*kgre_i915_gem_object_truncate)(obj);
	case __I915_MADV_PURGED:
		return;
	}

	if (obj->base.filp == NULL)
		return;

	mapping = file_inode(obj->base.filp)->i_mapping,
	invalidate_mapping_pages(mapping, 0, (loff_t)-1);
}

/* New. */
static struct mutex *kgrp_tlb_invalidate_lock;

/* New. */
#define kgrp__wait_for_us(COND, US, W) ({ \
	unsigned long timeout__ = jiffies + usecs_to_jiffies(US) + 1;   \
	int ret__;                                                      \
	for (;;) {                                                      \
		bool expired__ = time_after(jiffies, timeout__);        \
		if (COND) {                                             \
			ret__ = 0;                                      \
			break;                                          \
		}                                                       \
		if (expired__) {                                        \
			ret__ = -ETIMEDOUT;                             \
			break;                                          \
		}                                                       \
		usleep_range((W), (W)*2);                               \
	}                                                               \
	ret__;                                                          \
})

/* New. */
static int
kgrp___intel_wait_for_register_fw(struct drm_i915_private *dev_priv,
			     u32 reg,
			     const u32 mask,
			     const u32 value,
			     const unsigned int timeout_us,
			     const unsigned int timeout_ms)
{
#define done ((I915_READ_FW(reg) & mask) == value)
	int ret = kgrp__wait_for_us(done, timeout_us, 2);
	if (ret)
		ret = wait_for(done, timeout_ms);
	return ret;
#undef done
}

/* New. */
static void kgrp_invalidate_tlbs(struct drm_i915_private *dev_priv)
{
	static const u32 gen8_regs[] = {
		[RCS]  = KGRP_GEN8_RTCR,
		[VCS]  = KGRP_GEN8_M1TCR,
		[VCS2] = KGRP_GEN8_M2TCR,
		[VECS] = KGRP_GEN8_VTCR,
		[BCS]  = KGRP_GEN8_BTCR,
	};
	enum intel_ring_id id;

	if (INTEL_INFO(dev_priv)->gen < 8)
		return;

	mutex_lock(kgrp_tlb_invalidate_lock);
	(*kgre_intel_uncore_forcewake_get)(dev_priv, FORCEWAKE_ALL);

	for (id = 0; id < I915_NUM_RINGS; id++) {
		struct intel_engine_cs *engine = &dev_priv->ring[id];
		/*
		 * HW architecture suggest typical invalidation time at 40us,
		 * with pessimistic cases up to 100us and a recommendation to
		 * cap at 1ms. We go a bit higher just in case.
		 */
		const unsigned int timeout_us = 100;
		const unsigned int timeout_ms = 4;

		if (!(*kgre_intel_ring_initialized)(engine))
			continue;

		if (WARN_ON_ONCE(id >= ARRAY_SIZE(gen8_regs) || !gen8_regs[id]))
			continue;

		I915_WRITE_FW(gen8_regs[id], 1);
		if (kgrp___intel_wait_for_register_fw(dev_priv,
						 gen8_regs[id], 1, 0,
						 timeout_us, timeout_ms))
			KGRR_DRM_ERROR_RATELIMITED("%s TLB invalidation did not complete in %ums!\n",
					      engine->name, timeout_ms);
	}

	(*kgre_intel_uncore_forcewake_put)(dev_priv, FORCEWAKE_ALL);
	mutex_unlock(kgrp_tlb_invalidate_lock);
}

int
kgrp_i915_gem_object_put_pages(struct drm_i915_gem_object *obj)
{
	const struct drm_i915_gem_object_ops *ops = obj->ops;

	if (obj->pages == NULL)
		return 0;

	if (obj->pages_pin_count)
		return -EBUSY;

	BUG_ON((*kgre_i915_gem_obj_bound_any)(obj));

	/* ->put_pages might need to allocate memory for the bit17 swizzle
	 * array, hence protect them from being reaped by removing them from gtt
	 * lists early. */
	list_del(&obj->global_list);

	/*
	 * Fix CVE-2022-0330
	 *  +7 lines
	 */
	if (kgrp_bo_test_and_clear_was_bound(obj)) {
		struct drm_i915_private *i915 = to_i915(obj->base.dev);

		(*kgre_intel_runtime_pm_get)(i915);
		kgrp_invalidate_tlbs(i915);
		(*kgre_intel_runtime_pm_put)(i915);
	}

	ops->put_pages(obj);
	obj->pages = NULL;

	kgrr_i915_gem_object_invalidate(obj);

	return 0;
}



#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kgraft.h>
#include "kgr_patch_bsc1195950.h"
#include "../kallsyms_relocs.h"
#include "../shadow.h"

struct kgr_bsc1195950_shared_state
{
	unsigned long refcount;
	struct mutex tlb_invalidate_lock;
};

static struct kgr_bsc1195950_shared_state *kgr_bsc1195950_shared_state;

#define KGR_BSC1195950_SHARED_STATE_ID KGR_SHADOW_ID(1195950, 0)

static int kgr_bsc1195950_init_shared_state(void *obj,
					    void *shadow_data,
					    void *ctor_dat)
{
	struct kgr_bsc1195950_shared_state *s = shadow_data;

	memset(s, 0, sizeof(*s));
	mutex_init(&s->tlb_invalidate_lock);

	return 0;
}

static void __kgr_bsc1195950_put_shared_state(void);

/* Must be called with module_mutex held. */
static int __kgr_bsc1195950_get_shared_state(void)
{
	kgr_bsc1195950_shared_state =
		kgr_shadow_get_or_alloc(NULL, KGR_BSC1195950_SHARED_STATE_ID,
					sizeof(*kgr_bsc1195950_shared_state),
					GFP_KERNEL,
					kgr_bsc1195950_init_shared_state, NULL);
	if (!kgr_bsc1195950_shared_state)
		return -ENOMEM;

	++kgr_bsc1195950_shared_state->refcount;

	kgrp_tlb_invalidate_lock =
		&kgr_bsc1195950_shared_state->tlb_invalidate_lock;

	return 0;
}

/* Must be called with module_mutex held. */
static void __kgr_bsc1195950_put_shared_state(void)
{
	--kgr_bsc1195950_shared_state->refcount;

	if (!kgr_bsc1195950_shared_state->refcount) {
		mutex_destroy(&kgr_bsc1195950_shared_state->tlb_invalidate_lock);
		kgr_shadow_free(NULL, KGR_BSC1195950_SHARED_STATE_ID, NULL);
	}

	kgr_bsc1195950_shared_state = NULL;
}

#define KGR_PATCHED_MODULE "i915"

static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "drm_err", (void *)&kgre_drm_err, "drm" },
	{ "i915_gem_obj_bound_any", (void *)&kgre_i915_gem_obj_bound_any,
	  "i915" },
	{ "i915_gem_object_truncate", (void *)&kgre_i915_gem_object_truncate,
	  "i915" },
	{ "intel_ring_initialized", (void *)&kgre_intel_ring_initialized,
	  "i915" },
	{ "intel_runtime_pm_get", (void *)&kgre_intel_runtime_pm_get, "i915" },
	{ "intel_runtime_pm_put", (void *)&kgre_intel_runtime_pm_put, "i915" },
	{ "intel_uncore_forcewake_get",
	  (void *)&kgre_intel_uncore_forcewake_get, "i915" },
	{ "intel_uncore_forcewake_put",
	  (void *)&kgre_intel_uncore_forcewake_put, "i915" },
};

static int kgr_patch_bsc1195950_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1195950_module_nb = {
	.notifier_call = kgr_patch_bsc1195950_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1195950_i915_gem_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	ret = __kgr_bsc1195950_get_shared_state();
	if (ret)
		goto out;
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = __kgr_resolve_kallsyms_relocs(kgr_funcs,
						    ARRAY_SIZE(kgr_funcs));
		if (ret) {
			__kgr_bsc1195950_put_shared_state();
			goto out;
		}
	}

	ret = register_module_notifier(&kgr_patch_bsc1195950_module_nb);
	if (ret)
		__kgr_bsc1195950_put_shared_state();
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1195950_i915_gem_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1195950_module_nb);
	mutex_lock(&module_mutex);
	__kgr_bsc1195950_put_shared_state();
	mutex_unlock(&module_mutex);
}

#endif /* IS_ENABLED(CONFIG_DRM_I915) */
