/*
 * bsc1195950_i915_gem_gtt
 *
 * Fix for CVE-2022-0330, bsc#1195950 (drivers/gpu/drm/i915/i915_gem_gtt.c part)
 *
 *  Copyright (c) 2022 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_DRM_I915)

#if !IS_MODULE(CONFIG_DRM_I915)
#error "Live patch supports only CONFIG_DRM_I915=m"
#endif

#include <linux/tracepoint.h>

/* klp-ccp: from include/linux/tracepoint.h */
#define KGRR___DECLARE_TRACE(name, proto, args, cond, data_proto, data_args) \
	static struct tracepoint (*kgre___tracepoint_##name);	\
	static inline void kgrr_trace_##name(proto)		\
	{							\
		if (unlikely(static_key_enabled(&(*kgre___tracepoint_##name).key))) \
			__DO_TRACE(&(*kgre___tracepoint_##name),		\
					TP_PROTO(data_proto),			\
					TP_ARGS(data_args),			\
					TP_CONDITION(cond),,);			\
		if (IS_ENABLED(CONFIG_LOCKDEP) && (cond)) {			\
			rcu_read_lock_sched_notrace();				\
			rcu_dereference_sched((*kgre___tracepoint_##name).funcs); \
			rcu_read_unlock_sched_notrace();			\
		}								\
	}

#define KGRR_DECLARE_TRACE(name, proto, args) \
	KGRR___DECLARE_TRACE(name, PARAMS(proto), PARAMS(args), \
			cpu_online(raw_smp_processor_id()), \
			PARAMS(void *__data, proto), \
			PARAMS(__data, args))

#define KGRR_DEFINE_EVENT(template, name, proto, args)		\
	KGRR_DECLARE_TRACE(name, PARAMS(proto), PARAMS(args))

/* klp-ccp: from drivers/gpu/drm/i915/i915_gem_gtt.c */
#include <linux/seq_file.h>
#include <linux/stop_machine.h>
#include <drm/drmP.h>
#include <drm/i915_drm.h>

#include "bsc1195950_common.h"

/* klp-ccp: from drivers/gpu/drm/i915/i915_trace.h */
#define VM_TO_TRACE_NAME(vm) \
	(i915_is_ggtt(vm) ? "G" : \
		      "P")

KGRR_DEFINE_EVENT(i915_va, i915_va_alloc,
	     TP_PROTO(struct i915_address_space *vm, u64 start, u64 length, const char *name),
	     TP_ARGS(vm, start, length, name)
);

/* klp-ccp: from drivers/gpu/drm/i915/i915_gem_gtt.c */
int kgrp_i915_vma_bind(struct i915_vma *vma, enum i915_cache_level cache_level,
		  u32 flags)
{
	int ret;
	u32 bind_flags;

	if (WARN_ON(flags == 0))
		return -EINVAL;

	bind_flags = 0;
	if (flags & PIN_GLOBAL)
		bind_flags |= GLOBAL_BIND;
	if (flags & PIN_USER)
		bind_flags |= LOCAL_BIND;

	if (flags & PIN_UPDATE)
		bind_flags |= vma->bound;
	else
		bind_flags &= ~vma->bound;

	if (bind_flags == 0)
		return 0;

	if (vma->bound == 0 && vma->vm->allocate_va_range) {
		kgrr_trace_i915_va_alloc(vma->vm,
				    vma->node.start,
				    vma->node.size,
				    VM_TO_TRACE_NAME(vma->vm));

		/* XXX: i915_vma_pin() will fix this +- hack */
		vma->pin_count++;
		ret = vma->vm->allocate_va_range(vma->vm,
						 vma->node.start,
						 vma->node.size);
		vma->pin_count--;
		if (ret)
			return ret;
	}

	ret = vma->vm->bind_vma(vma, cache_level, bind_flags);
	if (ret)
		return ret;

	vma->bound |= bind_flags;

	/*
	 * Fix CVE-2022-0330
	 *  +2 lines
	 */
	if (vma->obj)
		kgrp_bo_set_was_bound(vma->obj);

	return 0;
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "kgr_patch_bsc1195950.h"
#include "../kallsyms_relocs.h"

#define KGR_PATCHED_MODULE "i915"

static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "__tracepoint_i915_va_alloc",
	  (void *)&kgre___tracepoint_i915_va_alloc, "i915" },
};

static int kgr_patch_bsc1195950_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1195950_module_nb = {
	.notifier_call = kgr_patch_bsc1195950_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1195950_i915_gem_gtt_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = __kgr_resolve_kallsyms_relocs(kgr_funcs,
						    ARRAY_SIZE(kgr_funcs));
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1195950_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1195950_i915_gem_gtt_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1195950_module_nb);
}

#endif /* IS_ENABLED(CONFIG_DRM_I915) */
